"""
Vectorize() with support for decorating methods; for example::

    from scipy.stats import rv_continuous
    from scipy_ext import vectorize

    class dist(rv_continuous):
        @vectorize(excluded=('n',), otypes=(float,))
        def _cdf(self, x, n):
            if n < 5:
                return f(x)  # One expensive calculation.
            else:
                return g(x)  # A different expensive calculation.

"""
from numpy import arange, stack, vectorize as numpy_vectorize


class _vectorize(numpy_vectorize):
    """
    Method decorator, working just like `numpy.vectorize()`.
    """
    def __get__(self, instance, owner):
        # Vectorize stores the decorated function (former "unbound method")
        # as pyfunc. Bound method's __get__ returns the method itself.
        self.pyfunc = self.pyfunc.__get__(instance, owner)
        return self


def vectorize(*args, **kwargs):
    """
    Allows using `@vectorize` as well as `@vectorize()`.
    """
    if args and callable(args[0]):
        # Guessing the argument is the method.
        return _vectorize(args[0])
    else:
        # Wait for the second call.
        return lambda m: _vectorize(m, *args, **kwargs)


def varange(starts, count):
    """
    Vectorized `arange()` taking a sequence of starts and a count of elements.

    For example::

        >>> varange(1, 5)
        array([1, 2, 3, 4, 5])

        >>> varange((1, 3), 5)
        array([[1, 2, 3, 4, 5],
               [3, 4, 5, 6, 7]])
    """
    try:
        return stack(arange(s, s + count) for s in starts)
    except TypeError:
        return arange(starts, starts + count)



"""
Distributions of the Kolmogorov-Smirnov supremum statistic.

After doi:10.18637/jss.v008.i18 and doi:10.18637/jss.v039.i11.
"""
from fractions import Fraction
from math import factorial, floor

from numpy import (arange, dot, exp, fmax, fromfunction, identity, log, modf,
                   pi, sqrt, tri)
from scipy.special import gamma, gammaln, smirnov
from scipy.stats import rv_continuous

from .vect import varange, vectorize


class ks_unif_gen(rv_continuous):
    """
    Approximate Kolmogorov-Smirnov two-sided, one-sample, distribution-free
    statistic (the hypothesized distribution continuous and fully specified).
    """
    def _argcheck(self, samples):
        return samples > 0

    @vectorize(otypes=(float,))
    def _cdf(self, statistic, samples):
        # Some simple, exact cases (more in Ruben & Gambino).
        if statistic <= 1 / (2 * samples):
            return 0.
        if statistic >= 1:
            return 1.
        if statistic <= 1 / samples:
            t = 2 * statistic - 1 / samples
            return exp(gammaln(samples + 1) + samples * log(t))
        if statistic >= 1 - 1 / samples:
            return 1 - 2 * (1 - statistic) ** samples

        # For small sample counts we may use an exact method when needed.
        if samples < 150:
            # With samples = 150 the matrix calculation takes about 100 ms
            # on a ~3 GFLOPS/core processor.
            if samples * statistic ** 2 < 7:
                # For a small threshold the Durbin matrix will be small.
                return ks_unif_durbin_matrix(samples, statistic)
            else:
                # Double the one-sided probability; accurate when close to one.
                return 1 - 2 * smirnov(samples, statistic)

        # Further we need to make a compromise between speed and accuracy.
        if samples < 100000 and samples * statistic ** 1.5 < 1.4:
            # The cost of the matrix calculation should still be acceptable.
            return ks_unif_durbin_matrix(samples, statistic)
        else:
            # No options left, but to use an asymptotic approximation.
            return ks_unif_pelz_good(samples, statistic)

    @vectorize(otypes=(float,))
    def _sf(self, statistic, samples):
        if statistic >= 1:
            # Statistic greater than 1 results in a NaN from Cephes smirnov().
            return 0.
        if statistic >= 1 - 1 / samples:
            # The _cdf code can suffer from some cancellation in this case.
            return min(1., 2 * (1 - statistic) ** samples)
        probability = 1 - self._cdf(statistic, samples)
        if probability > 1e-5:
            # Not too much precision got lost to cancellation.
            return probability
        else:
            # When the cdf float is very close to one it does not have bits
            # of small enough magnitude to express its 1-complement properly.
            # Hence, an approximate direct sf calculation may be more precise.
            return min(1., 2 * smirnov(samples, statistic))

ks_unif = ks_unif_gen(a=0, name='ks-unif', shapes='samples')


# Some arbitrary constants used for externalizing float exponents.
shift = 512
factor = float(2 ** shift)
factorr = float(2 ** -shift)


def ks_unif_durbin_matrix(samples, statistic):
    """
    Calculates the probability that the statistic is less than the given value,
    using a fairly accurate implementation of the Durbin's matrix formula.

    Not an exact transliteration of the Marsaglia code, but using the same
    ideas. Assumes samples > 0. See: doi:10.18637/jss.v008.i18.
    """
    # Construct the Durbin matrix.
    h, k = modf(samples * statistic)
    h = 1 - h
    m = 2 * k + 1
    A = tri(m, k=1)
    hs = h ** arange(1, m + 1)
    A[:, 0] -= hs
    A[-1] -= hs[::-1]
    if h > .5:
        A[-1, 0] += (2 * h - 1) ** m
    A /= fromfunction(lambda i, j: gamma(fmax(1, i - j + 2)), (m, m))
    # Calculate A ** n, expressed as P * 2 ** eP to avoid overflows.
    P = identity(m)
    s = samples
    eA, eP = 0, 0
    while s != 1:
        s, b = divmod(s, 2)
        if b == 1:
            P = dot(P, A)
            eP += eA
            if P[k, k] > factor:
                P /= factor
                eP += shift
        A = dot(A, A)
        eA *= 2
        if A[k, k] > factor:
            A /= factor
            eA += shift
    P = dot(P, A)
    eP += eA
    # Calculate n! / n ** n * P[k, k].
    x = P[k, k]
    for i in arange(1, samples + 1):
        x *= i / samples
        if x < factorr:
            x *= factor
            eP -= shift
    return x * 2 ** eP


def ks_unif_durbin_recurrence_rational(samples, statistic):
    """
    Calculates the probability that the statistic is less than the given value,
    using Durbin's recurrence and employing the standard fractions module.

    This is a (hopefully) exact reference implementation, likely too slow for
    practical usage. The statistic should be given as a Fraction instance and
    the result is also a Fraction. See: doi:10.18637/jss.v026.i02.
    """
    t = statistic * samples
    ft1 = floor(t) + 1
    fmt1 = floor(-t) + 1
    fdt1 = floor(2 * t) + 1
    qs = [Fraction(i ** i, factorial(i)) for i in range(ft1)]
    qs.extend(Fraction(i ** i, factorial(i)) - 2 * t *
                    sum((t + j) ** (j - 1) / factorial(j) *
                        (i - t - j) ** (i - j) / factorial(i - j)
                        for j in range(i + fmt1))
              for i in range(ft1, fdt1))
    qs.extend(-sum((-1) ** j * (2 * t - j) ** j / factorial(j) * qs[i - j]
                   for j in range(1, fdt1))
              for i in range(fdt1, samples + 1))
    return qs[samples] * factorial(samples) / samples ** samples


# Constants from the Pelz-Good approximation.
hs2 = varange(.5, 21) ** 2
ehs2 = exp(-hs2)
is2 = varange(1, 21) ** 2
pi2 = pi ** 2
pi4 = pi2 ** 2
pi6 = pi2 * pi4
hpi1d2 = sqrt(pi / 2)


def ks_unif_pelz_good(samples, statistic):
    """
    Approximates the statistic distribution by a transformed Li-Chien formula.

    This ought to be a bit more accurate than using the Kolmogorov limit, but
    should only be used with large squared sample count times statistic.
    See: doi:10.18637/jss.v039.i11 and http://www.jstor.org/stable/2985019.
    """
    x = 1 / statistic
    r2 = 1 / samples
    rx = sqrt(r2) * x
    r2x = r2 * x
    r2x2 = r2x * x
    r4x = r2x * r2
    r4x2 = r2x2 * r2
    r4x3 = r2x2 * r2x
    r5x3 = r4x2 * rx
    r5x4 = r4x3 * rx
    r6x3 = r4x2 * r2x
    r7x5 = r5x4 * r2x
    r9x6 = r7x5 * r2x
    r11x8 = r9x6 * r2x2
    a1 = rx * (-r6x3 / 108 + r4x2 / 18 - r4x / 36 - r2x / 3 + r2 / 6 + 2)
    a2 = pi2 / 3 * r5x3 * (r4x3 / 8 - r2x2 * 5 / 12 - r2x * 4 / 45 + x + 1 / 6)
    a3 = pi4 / 9 * r7x5 * (-r4x3 / 6 + r2x2 / 4 + r2x * 53 / 90 - 1 / 2)
    a4 = pi6 / 108 * r11x8 * (r2x2 / 6 - 1)
    a5 = pi2 / 18 * r5x3 * (r2x / 2 - 1)
    a6 = -pi4 * r9x6 / 108
    w = -pi2 / 2 * r2x2
    return hpi1d2 * ((a1 + (a2 + (a3 + a4 * hs2) * hs2) * hs2) * exp(w * hs2) +
                     (a5 + a6 * is2) * is2 * exp(w * is2)).sum()



"""
Goodness-of-fit tests based on the empirical distribution function.

Interface is similar to `scipy.stats.kstest()`, but you may choose or provide
a statistic calculation routine and a matching distribution.

Three concrete tests are provided:
* `ks_test()` -- Kolmogorov-Smirnov supremum statistic; almost the same as
  `scipy.stats.kstest()` with `alternative='two-sided'`, but with (hopefully)
  somewhat more precise p-value calculation;
* `cvm_test()` -- Cramer-von Mises L2 statistic, with a rather crude estimation
  of the statistic distribution (but seemingly the best available);
* `ad_test()` -- Anderson-Darling statistic with a fair approximation of its
  distribution; unlike the "composite" `scipy.stats.anderson()` this one needs
  a fully specified hypothesized distribution.

Example::

    >>> from scipy.stats import uniform
    >>> from skgof import ks_test

    >>> ks_test((1, 2, 3), uniform(0, 4))
    GofResult(statistic=0.25, pvalue=0.97222...)

The result is accurate -- if we assume that the samples were drawn from the
specified distribution, then P(D <= 1/4) = 2/9 * (2 * 3/4 - 1)^3 = .97(2).
In general, for sample counts less than 150 you may expect good precision
with `ks_test()`, and a fair one above that.

Lectures 2 and 3 of http://www.win.tue.nl/~rmcastro/AppStat2013/ list formulas
for all three statistics. Their distributions are split into separate modules
as calculating each is a small research story -- see `ksdist`, `cvmdist`, and
`addist` for details and further references.
"""
from collections import namedtuple
from functools import partial

from numpy import arange, log, sort
from scipy._lib.six import string_types
from scipy.stats import distributions

from .addist import ad_unif
from .cvmdist import cvm_unif
from .ksdist import ks_unif

GofResult = namedtuple('GofResult', ('statistic', 'pvalue'))


def ks_stat(data):
    """
    Calculates the Kolmogorov-Smirnov statistic for sorted values from U(0, 1).
    """
    samples = len(data)
    uniform = arange(0, samples + 1) / samples
    d_plus = (uniform[1:] - data).max()
    d_minus = (data - uniform[:-1]).max()
    return max(d_plus, d_minus)


def cvm_stat(data):
    """
    Calculates the Cramer-von Mises statistic for sorted values from U(0, 1).
    """
    samples2 = 2 * len(data)
    minuends = arange(1, samples2, 2) / samples2
    return 1 / (6 * samples2) + ((minuends - data) ** 2).sum()


def ad_stat(data):
    """
    Calculates the Anderson-Darling statistic for sorted values from U(0, 1).

    The statistic is not defined if any of the values is exactly 0 or 1. You
    will get infinity as a result and a divide-by-zero warning for such values.
    The warning can be silenced or raised using numpy.errstate(divide=...).
    """
    samples = len(data)
    factors = arange(1, 2 * samples, 2)
    return -samples - (factors * log(data * (1 - data[::-1]))).sum() / samples


def ecdfgof_test(data, dist, args=(), stat=ad_stat, pdist=ad_unif):
    """
    Tests goodness of fit of data to dist using a distribution-free statistic.
    """
    if isinstance(data, string_types):
        # Auto-generating samples from a named distribution is not supported.
        raise AttributeError("Data should be an array or list of values.")
    if isinstance(dist, string_types):
        dist = getattr(distributions, dist)(*args)
    elif args:
        dist = dist(args)
    statistic = stat(dist.cdf(sort(data)))
    pvalue = pdist(len(data)).sf(statistic)
    return GofResult(statistic, pvalue)

ks_test = partial(ecdfgof_test, stat=ks_stat, pdist=ks_unif)
cvm_test = partial(ecdfgof_test, stat=cvm_stat, pdist=cvm_unif)
ad_test = partial(ecdfgof_test, stat=ad_stat, pdist=ad_unif)



"""
Distributions of the Anderson-Darling statistic.

After doi:18637/jss.v009.i02.
"""
from numpy import exp, log, sqrt
from scipy.stats import rv_continuous

from .vect import vectorize


class ad_unif_gen(rv_continuous):
    """
    Approximate distribution of the uniform Anderson-Darling statistic
    (with the hypothesized distribution continuous and fully specified).
    """
    def _argcheck(self, samples):
        return samples > 0

    @vectorize(otypes=(float,))
    def _cdf(self, statistic, samples):
        if samples == 1:
            # Exact distribution for a single sample (a bit more precise than
            # the approximation). See doi:10.1214/aoms/1177704850 equation 8.
            if statistic <= log(4) - 1:
                return 0.
            else:
                return sqrt(1 - 4 * exp(-1 - statistic))
        pinf = ad_unif_inf(statistic)
        return pinf + ad_unif_fix(samples, pinf)

ad_unif = ad_unif_gen(a=0, name='ad-unif', shapes='samples')


def ad_unif_inf(statistic):
    """
    Approximates the limiting distribution to about 5 decimal digits.
    """
    z = statistic
    if z < 2:
        return (exp(-1.2337141 / z) / sqrt(z) *
                                (2.00012 + (.247105 - (.0649821 - (.0347962 -
                                (.011672 - .00168691 * z) * z) * z) * z) * z))
    else:
        return exp(-exp(1.0776 - (2.30695 - (.43424 - (.082433 -
                                (.008056 - .0003146 * z) * z) * z) * z) * z))


g1 = lambda x: sqrt(x) * (1 - x) * (49 * x - 102)
g2 = lambda x: (-.00022633 + (6.54034 - (14.6538 - (14.458 -
                                (8.259 - 1.91864 * x) * x) * x) * x) * x)
g3 = lambda x: (-130.2137 + (745.2337 - (1705.091 - (1950.646 -
                                (1116.36 - 255.7844 * x) * x) * x) * x) * x)


def ad_unif_fix(samples, pinf):
    """
    Corrects the limiting distribution for a finite sample size.
    """
    n = samples
    c = .01265 + .1757 / n
    if pinf < c:
        return (((.0037 / n + .00078) / n + .00006) / n) * g1(pinf / c)
    elif pinf < .8:
        return ((.01365 / n + .04213) / n) * g2((pinf - c) / (.8 - c))
    else:
        return g3(pinf) / n



"""
Distributions of the Cramer-von Mises statistic.

After doi:10.2307/2346175.
"""
from numpy import arange, dot, exp, newaxis, pi, tensordot
from scipy.special import gamma, kv
from scipy.stats import rv_continuous

from .vect import varange, vectorize


class cvm_unif_gen(rv_continuous):
    """
    Approximate Cramer-von Mises statistic distribution for uniform data
    (with the hypothesized distribution continuous and fully specified).
    """
    def _argcheck(self, samples):
        return samples > 0

    @vectorize(otypes=(float,))
    def _cdf(self, statistic, samples):
        low = 1 / (12 * samples)
        # Basic bounds.
        if statistic <= low:
            return 0.
        if statistic >= samples / 3:
            return 1.
        # From the geometric approach of Csorgo and Faraway (equation 2.4).
        if statistic <= low + 1 / (4 * samples ** 2):
            return (gamma(samples + 1) / gamma(samples / 2 + 1) *
                                    (pi * (statistic - low)) ** (samples / 2))
        # Asymptotic distribution with a one-term correction (equation 1.8).
        return cvm_unif_inf(statistic) + cvm_unif_fix1(statistic) / samples

cvm_unif = cvm_unif_gen(a=0, name='cvm-unif', shapes='samples')


inf_ks41 = 4 * arange(11) + 1
inf_args = inf_ks41 ** 2 / 16
inf_cs = (inf_ks41 ** .5 * gamma(varange(.5, 11)) /
                                        (pi ** 1.5 * gamma(varange(1, 11))))


def cvm_unif_inf(statistic):
    """
    Calculates the limiting distribution of the Cramer-von Mises statistic.

    After the second line of equation 1.3 from the Csorgo and Faraway paper.
    """
    args = inf_args / statistic
    return (inf_cs * exp(-args) * kv(.25, args)).sum() / statistic ** .5


fix1_args = (4 * (varange((.5, 1., 1.5), 21)) - 1) ** 2 / 16
fix1_dens = 72 * pi ** 1.5 * gamma(varange(1, 21))
fix1_csa = fix1_args ** .75 * gamma(varange(1.5, 21)) / fix1_dens
fix1_csb = fix1_args ** 1.25 * gamma(varange((.5, 1.5, 2.5), 21)) / fix1_dens


def cvm_unif_fix1(statistic):
    """
    Approximates the first-term of the small sample count Gotze expansion.

    After equation 1.10 (with coefficients pulled out as csa / csb).
    """
    args = fix1_args / statistic
    kvs = kv((.25, .75, 1.25), args[:, :, newaxis])
    gs, hs = exp(-args) * tensordot(((1, 1, 0), (2, 3, -1)), kvs, axes=(1, 2))
    a = dot((7, 16, 7), fix1_csa * gs).sum() / statistic ** 1.5
    b = dot((1, 0, 24), fix1_csb * hs).sum() / statistic ** 2.5
    return cvm_unif_inf(statistic) / 12 - a - b



#!/usr/bin/env python

from setuptools import setup

setup(
    name='scikit-gof',
    version='0.0.2',
    packages=('skgof',),
    install_requires=('numpy>=1.10', 'scipy'),
    tests_require=('pytest', 'pytest-flake8', 'pytest-isort', 'pytest-readme',
                   'flake8-print', 'flake8-todo', 'pep8-naming'),

    description="Variations on goodness of fit tests for SciPy.",
    long_description=open('README.rst').read(),
    author="Wojciech Ruszczewski",
    author_email="scipy@wr.waw.pl",
    url="http://github.org/wrwrwr/scikit-gof",
    license="MIT",
    classifiers=(
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.4",
    ),
)



"""
A primitive statistic distribution simulation.
"""
from numpy import fromiter
from numpy.random import random_sample


def simulator(stat, samples, precision, rounds):
    """
    Simulates a distribution-free statistical test to estimate its p-values.

    The first argument, statistic, should be a function that computes the
    simulated test for a vector of ordered, uniform(0, 1) samples.

    The second argument, samples, tells how many samples to generate and test
    with. To generate a typical p-values table you would run the function for
    a number of consecutive sample counts.

    The third argument, precision, decides how many critical values to return.
    For example with precision = 100 you will get 99 values; approximately, a
    statistic above value indexed 94 will have a probability lower than .05.

    The fourth argument, rounds, tells us how many times to repeat the data
    generation and statistic calculation. The more rounds the higher the
    quality of the results. Must be a (large) multiple of precision.

    Example::

        from skgof.testsim import simulator
        from skgof.ecdfgof import ks_stat

        # Repeat the simulation one million times for vectors of 10 samples.
        ks10 = simulator(ks_stat, 10, 100, 1000000)

        # Get the approximate 95% critical value (to about 2 decimal digits).
        ks10[94]  # 0.409...

    """
    data = random_sample(size=(rounds, samples))
    data.sort(axis=1)
    stats = fromiter((stat(d) for d in data), float, rounds)
    stats.sort()
    step = int(rounds / precision)
    return stats[step:rounds:step]
